from datetime import (
    timedelta
)


from django.conf import (
    settings,
)
from django.core.management.base import (
    BaseCommand,
)
from django.utils import (
    timezone,
)


from ...models import (
    PostCategory,
    PostInstance,
)


class Command(BaseCommand):
    help = 'Deletes form responses that have reached or passed their deletion date'

    def handle(self, *args, **options):
        try:
            days_saved = settings.FORMS_DELETE_RESPONSES
        except AttributeError:
            days_saved = 365
        today = timezone.now().date()
        expiration_date = today - timedelta(days=days_saved)
        PostInstance.objects.filter(
            created__date__lte=expiration_date,
        ).delete()
        PostCategory.objects.filter(
            form_instance=None,
        ). filter(
            post_instances=None,
        ).delete()



import copy
from datetime import (
    date,
    datetime,
    time,
)


from django.contrib.sites.models import (
    Site,
)
from django.db import (
    models,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone,
)
from django.utils.crypto import (
    get_random_string,
)
from django.utils.formats import (
    localize,
)
from django.utils.text import (
    slugify,
)


from ckeditor.fields import (
    RichTextField,
)
from django_simple_file_handler.models import (
    PrivatePDF,
)


class BaseMixin(models.Model):
    created = models.DateTimeField(
        auto_now_add=True,
    )
    updated = models.DateTimeField(
        'last updated',
        auto_now=True,
    )

    class Meta:
        abstract = True


class FormsSiteProfile(BaseMixin, Site):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    protocol = models.CharField(
        max_length=255,
        default='https://',
    )

    def protocol_domain(self):
        return f'{self.protocol}{self.domain}'

    class Meta:
        verbose_name = 'site profile'


class EmailAddress(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    recipient_name = models.CharField(
        max_length=255,
        default='Anonymous',
    )
    email = models.EmailField(
    )

    def __str__(self):
        return self.email

    def inst_count(self):
        count = self.form_instances.all().count()
        return str(count)

    inst_count.short_description = 'forms'

    class Meta:
        verbose_name_plural = 'email addresses'


class DictDefinition(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    input_definition = models.ForeignKey(
        'InputDefinition',
        on_delete=models.CASCADE,
    )
    value = models.CharField(
        max_length=255,
    )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.input_definition.downstream_update()

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        self.input_definition.downstream_update()

    class Meta:
        abstract = True


ERROR_CHOICES = [
    ('required', 'required'),
    ('invalid', 'invalid (Datefield, EmailField, TimeField only)'),
    ('invalid_choice', 'invalid_choice (ChoiceField, MultipleChoiceField only)'),
    ('invalid_list', 'invalid_list (MultipleChoiceField only)'),
    ('max_length', 'max_length (CharField only)'),
    ('min_length', 'min_length (CharField only)'),
]


class ErrorDefinition(DictDefinition):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    key = models.CharField(
        max_length=255,
        choices=ERROR_CHOICES,
        default='required'
    )

    def __str__(self):
        return f'{self.key}: {self.value}'

    class Meta:
        verbose_name = 'error message'


class ChoiceDefinition(DictDefinition):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    display = models.CharField(
        max_length=255,
    )
    sort_order = models.PositiveIntegerField(
        default=0,
        blank=False,
        null=False
    )

    def __str__(self):
        return f'{self.value}: {self.display}'

    class Meta(object):
        ordering = [
            'sort_order',
        ]
        verbose_name = 'choice'
        verbose_name_plural = 'choices (ChoiceField and MultipleChoiceField only'


FIELD_TYPES = [
    ('BooleanField', 'BooleanField'),
    ('CharField', 'CharField'),
    ('ChoiceField', 'ChoiceField'),
    ('DateField', 'DateField'),
    ('EmailField', 'EmailField'),
    ('MultipleChoiceField', 'MultipleChoiceField'),
    ('TimeField', 'TimeField'),
]


WIDGET_TYPES = [
    ('CheckboxInput', 'CheckboxInput'),
    ('CheckboxSelectMultiple', 'CheckboxSelectMultiple'),
    ('DateInput', 'DateInput'),
    ('EmailInput', 'EmailInput'),
    ('Select', 'Select'),
    ('SelectMultiple', 'SelectMultiple'),
    ('Textarea', 'Textarea'),
    ('TextInput', 'TextInput'),
    ('TimeInput', 'TimeInput'),
]


class InputDefinition(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    definition_name = models.CharField(
        max_length=255,
        unique=True,
    )
    field_type = models.CharField(
        max_length=255,
        choices=FIELD_TYPES,
        default='CharField',
    )
    label = models.CharField(
        max_length=255,
    )
    label_suffix = models.CharField(
        max_length=255,
        default=':',
    )
    min_length = models.PositiveIntegerField(
        'min length (CharField and EmailField only)',
        blank=True,
        null=True,
    )
    max_length = models.PositiveIntegerField(
        'max length (CharField and EmailField only)',
        default=255,
        blank=True,
        null=True,
    )
    initial = models.CharField(
        'initial value',
        max_length=255,
        blank=True,
    )
    initial_boolean = models.BooleanField(
        'initial True (BooleanField only)',
        default=False,
    )
    boolean_true = models.CharField(
        'boolean true string (if select widget)',
        max_length=255,
        blank=True,
    )
    boolean_false = models.CharField(
        'boolean false string (if select widget)',
        max_length=255,
        blank=True,
    )
    empty_value = models.CharField(
        'empty value (CharField only)',
        max_length=255,
        blank=True,
    )
    help_text = RichTextField(
        blank=True,
    )
    strip = models.BooleanField(
        'strip (CharField only)',
        default=True,
    )
    required = models.BooleanField(
        default=True,
    )
    widget_type = models.CharField(
        max_length=255,
        choices=WIDGET_TYPES,
        default='TextInput',
    )
    widget_autocomplete = models.BooleanField(
        'autocomplete',
        default=True,
    )
    widget_title = models.CharField(
        'title',
        max_length=255,
        blank=True,
    )
    widget_class = models.CharField(
        'class',
        max_length=255,
        blank=True,
    )

    def __str__(self):
        return self.definition_name

    def full_label(self):
        return f'{self.label}{self.label_suffix}'

    def dict_data(self):
        field_data = {
            'field_type': self.field_type,
            'field_args': {
                'label': self.label,
                'label_suffix': self.label_suffix,
                'required': self.required,
            }
        }
        field_args = field_data['field_args']
        if self.help_text:
            field_args['help_text'] = self.help_text
        error_messages = {
            'required': 'Field is required',
            'invalid': 'Invalid value',
            'invalid_choice': 'Invalid choice',
            'invalid_list': 'Invalid list',
            'max_length': 'Maximum length exceeded',
            'min_length': 'Minimum length not met',
        }
        error_definitions = ErrorDefinition.objects.filter(
            input_definition=self,
        )
        if error_definitions:
            for item in error_definitions.all():
                error_messages[item.key] = item.value
        field_args['error_messages'] = error_messages
        if self.field_type == 'BooleanField':
            if self.widget_type == 'Select':
                field_data['field_type'] = 'TypedChoiceField'
                field_args['boolean_dicts'] = [{False: self.boolean_false}, {True: self.boolean_true}]
                if self.initial_boolean:
                    field_args['initial'] = 'true'
                else:
                    field_args['initial'] = 'false'
            else:
                if self.initial_boolean:
                    field_args['initial'] = True
        else:
            if self.initial:
                field_args['initial'] = self.initial
        if self.field_type == 'CharField' or self.field_type == 'EmailField':
            if self.min_length:
                field_args['min_length'] = self.min_length
            if self.max_length:
                field_args['max_length'] = self.max_length
            if self.field_type == 'CharField':
                field_args['strip'] = self.strip
                field_args['empty_value'] = self.empty_value
        if self.field_type == 'ChoiceField' or self.field_type == 'MultipleChoiceField':
            choice_definitions = ChoiceDefinition.objects.filter(
                input_definition=self,
            )
            choice_dicts = [{item.value: item.display} for item in choice_definitions]
            field_args['choice_dicts'] = choice_dicts
        widget_data = {
            'widget_type': self.widget_type,
            'widget_attrs': {
            },
        }
        widget_attrs = widget_data['widget_attrs']      # Check which of these work for which field type
        if not self.widget_autocomplete:
            widget_attrs['autocomplete'] = 'off'
        if self.widget_title:
            widget_attrs['title'] = self.widget_title
        if self.widget_type == 'CheckboxSelectMultiple':
            default_class = 'checkbox_select_multiple'
        else:
            default_class = ''
        if default_class and self.widget_class:
            class_space = ' '
        else:
            class_space = ''
        class_str = f'{default_class}{class_space}{self.widget_class}'
        if class_str:
            widget_attrs['class'] = class_str
        data = {
            'field_data': field_data,
            'widget_data': widget_data,
        }
        return data

    def inst_count(self):
        count = self.input_instances.all().count()
        return str(count)

    inst_count.short_description = 'instances'

    def downstream_update(self, def_delete=False):
        for instance in self.input_instances.all():
            instance.downstream_update(
                def_delete=def_delete,
            )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.downstream_update()

    def delete(self, *args, **kwargs):
        self.downstream_update(
            def_delete=True,
        )
        super().delete(*args, **kwargs)

    class Meta:
        ordering = [
            'definition_name',
        ]


class InputInstance(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    form_definition = models.ForeignKey(
        'FormDefinition',
        on_delete=models.CASCADE,
        related_name='input_instances',
    )
    input_definition = models.ForeignKey(
        'InputDefinition',
        on_delete=models.CASCADE,
        related_name='input_instances',
    )
    field_name = models.CharField(
        max_length=255,
    )
    admin_only = models.BooleanField(
        default=False,
    )
    sort_order = models.PositiveIntegerField(
        default=0,
        blank=False,
        null=False
    )

    def __str__(self):
        return self.field_name

    def dict_data(self):
        if self.input_definition:
            data = self.input_definition.dict_data()
            data['field_data']['field_name'] = self.field_name
            data['admin_only'] = self.admin_only
            return data

    def downstream_update(self, def_delete=False):
        if def_delete:
            self.input_definition = None
            self.delete()
        self.form_definition.downstream_update()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.downstream_update()

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        self.downstream_update()

    class Meta(object):
        ordering = [
            'sort_order',
        ]
        verbose_name = 'input'


class TableField(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    form_definition = models.ForeignKey(
        'FormDefinition',
        on_delete=models.CASCADE,
        related_name='table_fields',
    )
    input_instance = models.OneToOneField(
        'InputInstance',
        on_delete=models.CASCADE,
        related_name='table_fields',
    )
    column_name = models.CharField(
        'custom column header',
        max_length=255,
        blank=True,
    )
    tally_column = models.BooleanField(
        'tally column (BooleanField only)',
        default=False,
    )
    sort_order = models.PositiveIntegerField(
        default=0,
        blank=False,
        null=False
    )

    class Meta(object):
        ordering = [
            'sort_order',
        ]


class FormDefinition(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.from_admin = False

    definition_name = models.CharField(
        max_length=255,
        unique=True,
    )
    send_email = models.BooleanField(
        default=True,
    )
    subject_prefix = models.CharField(
        'subject prefix string',
        max_length=255,
        blank=True,
        default='Form email',
    )
    email_html = models.CharField(
        'HTML email template',
        max_length=255,
        default='django_simple_forms/internal/emails/form_email.html',
    )
    email_txt = models.CharField(
        'text email template',
        max_length=255,
        default='django_simple_forms/internal/emails/form_email.txt',
    )
    reply_first = models.CharField(
        'reply-to first name CharField',
        max_length=255,
        blank=True,
    )
    reply_last = models.CharField(
        'reply-to last name CharField',
        max_length=255,
        blank=True,
    )
    reply_address = models.CharField(
        'reply-to email EmailField',
        max_length=255,
        blank=True,
    )
    email_subject = models.CharField(
        'email subject CharField',
        max_length=255,
        blank=True,
    )
    create_pdf = models.BooleanField(
        'create PDF',
        default=False,
    )
    pdf_template = models.CharField(
        'PDF template',
        max_length=255,
        default='django_simple_forms/internal/pdfs/default_pdf.html',
    )
    editable_data = models.BooleanField(
        default=False,
    )
    sort_created = models.BooleanField(
        'sort by date created (if no table fields specified)',
        default=True,
    )
    form_html = models.CharField(
        'form HTML (external)',
        max_length=255,
        default='django_simple_forms/external/pages/form_page.html',
    )
    form_css = models.CharField(
        'form CSS (external)',
        max_length=255,
        default='django_simple_forms/external/css/form_style.css',
    )
    success_html = models.CharField(
        'success page HTML (external)',
        max_length=255,
        default='django_simple_forms/external/pages/form_submitted.html',
    )
    button_label = models.CharField(
        max_length=255,
        default='Submit',
    )
    post_html = models.CharField(
        'response HTML (internal)',
        max_length=255,
        default='django_simple_forms/internal/pages/post_view.html',
    )
    post_form_html = models.CharField(
        'form HTML (internal)',
        max_length=255,
        default='django_simple_forms/internal/pages/post_form.html',
    )

    def __str__(self):
        return self.definition_name

    def inst_count(self):
        count = self.form_instances.all().count()
        return str(count)

    inst_count.short_description = 'instances'

    def table_header(self, field):
        if field.column_name:
            header = field.column_name
        else:
            header = field.input_instance.input_definition.full_label()
        return header

    def dict_data(self):
        data = {
            'field_dicts': [instance.dict_data() for instance in self.input_instances.all()],
            'table_fields': [{field_table.input_instance.field_name: self.table_header(field_table)}
                             for field_table in self.table_fields.all()],
            'tally_fields': [field_tally.input_instance.field_name
                             for field_tally in self.table_fields.all()
                             if field_tally.tally_column
                             and field_tally.input_instance.input_definition.field_type == 'BooleanField'],
            'sort_created': self.sort_created,
            'date_fields': [field_date.field_name
                            for field_date in self.input_instances.all()
                            if field_date.input_definition.field_type == 'DateField'],
            'time_fields': [field_time.field_name
                            for field_time in self.input_instances.all()
                            if field_time.input_definition.field_type == 'TimeField'],
        }
        return data

    def downstream_update(self):
        for instance in self.form_instances.all():
            instance.save()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if not self.from_admin:
            self.downstream_update()


class FormInstance(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    title = models.CharField(
        max_length=255,
        unique=True,
    )
    alt_title = models.CharField(
        max_length=255,
        blank=True,
    )
    form_definition = models.ForeignKey(
        'FormDefinition',
        on_delete=models.CASCADE,
        related_name='form_instances',
    )
    published = models.BooleanField(
        default=False,
    )
    descriptive_text = RichTextField(
        blank=True,
    )
    email_addresses = models.ManyToManyField(
        EmailAddress,
        related_name='form_instances',
        blank=True,
    )
    data_json = models.JSONField(
        blank=True,
        null=True,
    )
    footer_text = RichTextField(
        blank=True,
    )

    def __str__(self):
        return self.title

    def form_url(self):
        url_str = reverse(
            'django_simple_forms:form_page',
            kwargs={
                'pk': self.pk,
                'title_slug': slugify(self.title),
            },
        )
        return url_str

    def dict_data(self):
        return self.data_json

    def save(self, *args, **kwargs):
        self.data_json = self.form_definition.dict_data()
        super().save(*args, **kwargs)
        post_category, created = PostCategory.objects.get_or_create(
            form_instance=self,
        )
        post_category.save()

    class Meta:
        ordering = [
            'title',
        ]


class PostCategory(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    category = models.CharField(
        max_length=255,
        default='Uncategorized',
    )
    form_instance = models.OneToOneField(
        'FormInstance',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='post_category',
    )
    data_json = models.JSONField(
        blank=True,
        null=True,
    )

    def __str__(self):
        return self.category

    def inst_count(self):
        count = self.post_instances.all().count()
        return str(count)

    inst_count.short_description = 'response count'

    def save(self, *args, **kwargs):
        if self.form_instance:
            self.category = self.form_instance.title
            self.data_json = {
                'table_fields': self.form_instance.data_json['table_fields'],
                'tally_fields': self.form_instance.data_json['tally_fields'],
                'sort_created': self.form_instance.data_json['sort_created'],
                'date_fields': self.form_instance.data_json['date_fields'],
                'time_fields': self.form_instance.data_json['time_fields'],
            }
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        for post in self.post_instances.all():
            post.delete()
        if not self.form_instance:
            super().delete(*args, **kwargs)

    class Meta:
        verbose_name = 'form responses'
        verbose_name_plural = 'form responses'


class PostPDF(PrivatePDF):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    class Meta:
        proxy = True
        verbose_name = 'post PDF'


def format_data(form_instance=None, cleaned_data=None):
    if form_instance and cleaned_data:
        for key, value in cleaned_data.items():
            if value is None:
                cleaned_data[key] = ''
            if isinstance(value, date):
                cleaned_data[key] = value.strftime('%Y-%m-%d')
            if isinstance(value, time):
                cleaned_data[key] = value.strftime('%H:%M:%S')
            if isinstance(value, str):
                cleaned_data[key] = value.replace('\r\n', '<br>')
            if isinstance(value, list):
                formatted_list = []
                for data in cleaned_data[key]:
                    if isinstance(data, str):
                        data.replace('\r\n', '<br>')
                    formatted_list.append(data)
                cleaned_data[key] = formatted_list
        label_dicts = [
            {item['field_data']['field_name']:
                item['field_data']['field_args']['label'] + item['field_data']['field_args']['label_suffix']}
            for item in form_instance.dict_data()['field_dicts']
        ]
        stored_dict = {'data_dicts': []}
        for pair in label_dicts:
            for key, value in pair.items():
                data_value = cleaned_data[key]
                if isinstance(data_value, list):
                    value_list = [value] + data_value
                    stored_dict['data_dicts'].append({key: value_list})
                else:
                    stored_dict['data_dicts'].append({key: [value, data_value]})
    else:
        stored_dict = {}
    return stored_dict


class PostInstanceManager(models.Manager):
    def create_instance(self, form_instance=None, cleaned_data=None):
        stored_dict = format_data(
            form_instance=form_instance,
            cleaned_data=cleaned_data,
        )
        post = self.create(
            post_category=form_instance.post_category,
            data_json=stored_dict,
        )
        return post


class PostInstance(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    post_category = models.ForeignKey(
        'PostCategory',
        on_delete=models.CASCADE,
        related_name='post_instances',
    )
    data_json = models.JSONField(
        blank=True,
        null=True,
    )
    pdf = models.OneToOneField(
        'PostPDF',
        on_delete=models.SET_NULL,
        related_name='post_instance',
        blank=True,
        null=True,
    )
    objects = PostInstanceManager()

    def output_dict(self):
        output = {}
        if self.data_json:
            try:
                data_dicts = self.data_json['data_dicts']
                for pair in data_dicts:
                    for key, value in pair.items():
                        output[key] = value
            except KeyError:
                pass
        return output

    def output_no_br(self):
        output_copy = copy.deepcopy(self.output_dict())
        for key, value in output_copy.items():
            unformatted_list = [item.replace('<br>', '\r\n') if isinstance(item, str) else item for item in value]
            output_copy[key] = unformatted_list
        return output_copy

    def output_formatted_date_time(self):
        output_dict_copy = copy.deepcopy(self.output_dict())
        for date_field in self.data_json['date_fields']:
            try:
                date_in = output_dict_copy[date_field][1]
                date_obj = datetime.strptime(date_in, '%Y-%m-%d')
                date_out = localize(date_obj.date())
                output_dict_copy[date_field][1] = date_out
            except (KeyError, ValueError) as error:
                pass
        for time_field in self.data_json['time_fields']:
            try:
                time_in = output_dict_copy[time_field][1]
                time_obj = datetime.strptime(time_in, '%H:%M:%S')
                time_out = localize(time_obj.time())
                output_dict_copy[time_field][1] = time_out
            except (KeyError, ValueError) as error:
                pass
        return output_dict_copy

    def output_text(self):
        output_formatted_date_time = copy.deepcopy(self.output_formatted_date_time())
        for key, value in output_formatted_date_time.items():
            unformatted_list = [item.replace('<br>', '\r\n') if isinstance(item, str) else item for item in value]
            output_formatted_date_time[key] = unformatted_list
        return output_formatted_date_time

    def save(self, *args, **kwargs):
        if not self.data_json:
            self.data_json = {}
        self.data_json['date_fields'] = self.post_category.data_json['date_fields']
        self.data_json['time_fields'] = self.post_category.data_json['time_fields']
        if self.post_category.form_instance:
            if self.post_category.form_instance.form_definition.create_pdf:
                if self.pdf:
                    self.pdf.delete()
                title = self.post_category.form_instance.title
                file_title = f'{title}-{get_random_string(20)}'
                local_time = timezone.localtime(timezone.now())
                if not self.created:
                    self.created = local_time
                    self.updated = local_time
                else:
                    self.updated = local_time
                self.pdf = PostPDF.objects.create(
                    title=file_title,
                    template_location=self.post_category.form_instance.form_definition.pdf_template,
                    template_data={
                        'title': title,
                        'created': self.created,
                        'updated': self.updated,
                        'descriptive_text': self.post_category.form_instance.descriptive_text,
                        'footer_text': self.post_category.form_instance.footer_text,
                        'output_dict': self.output_dict(),
                        'output_formatted_date_time': self.output_formatted_date_time(),
                    },
                )
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        if self.pdf:
            self.pdf.delete()
        super().delete(*args, **kwargs)

    class Meta:
        ordering = [
            '-created',
        ]



from django.urls import (
    path,
)


from .views import (
    clone_email_address,
    clone_form_definition,
    clone_form_instance,
    clone_input_definition,
    clone_post_instance,
    form_page,
    form_page_admin,
    form_page_admin_add,
    post_view,
    post_spreadsheet,
)


app_name = 'django_simple_forms'


urlpatterns = [
    path(
        '<pk>/<title_slug>.html',
        form_page,
        name='form_page',
    ),
    path(
        'clone-email-address/<app_label>/<model_name>/<pk>/',
        clone_email_address,
        name='clone_email_address',
    ),
    path(
        'clone-form-definition/<app_label>/<model_name>/<pk>/',
        clone_form_definition,
        name='clone_form_definition',
    ),
    path(
        'clone-form-instance/<app_label>/<model_name>/<pk>/',
        clone_form_instance,
        name='clone_form_instance',
    ),
    path(
        'clone-input-definition/<app_label>/<model_name>/<pk>/',
        clone_input_definition,
        name='clone_input_definition',
    ),
    path(
        'clone-post-instance/<app_label>/<model_name>/<pk>/',
        clone_post_instance,
        name='clone_post_instance',
    ),
    path(
        'post-add/<app_label>/<model_name>/<pk>/add/',
        form_page_admin_add,
        name='form_page_admin_add',
    ),
    path(
        'post-edit/<app_label>/<model_name>/<pk>/edit/',
        form_page_admin,
        name='form_page_admin',
    ),
    path(
        'post-view/<app_label>/<model_name>/<pk>/view/',
        post_view,
        name='post_view',
    ),
    path(
        'post-spreadsheet/<app_label>/<model_name>/<pk>/',
        post_spreadsheet,
        name='post_spreadsheet',
    ),
]



import ckeditor.fields
import django.contrib.sites.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('django_simple_file_handler', '0008_auto_20210702_1604'),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailAddress',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('recipient_name', models.CharField(default='Anonymous', max_length=255)),
                ('email', models.EmailField(max_length=254)),
            ],
            options={
                'verbose_name_plural': 'email addresses',
            },
        ),
        migrations.CreateModel(
            name='FormDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('definition_name', models.CharField(max_length=255, unique=True)),
                ('send_email', models.BooleanField(default=True)),
                ('subject_prefix', models.CharField(blank=True, default='Form email', max_length=255, verbose_name='subject prefix string')),
                ('email_html', models.CharField(default='django_simple_forms/internal/emails/form_email.html', max_length=255, verbose_name='HTML email template')),
                ('email_txt', models.CharField(default='django_simple_forms/internal/emails/form_email.txt', max_length=255, verbose_name='text email template')),
                ('reply_first', models.CharField(blank=True, max_length=255, verbose_name='reply-to first name CharField')),
                ('reply_last', models.CharField(blank=True, max_length=255, verbose_name='reply-to last name CharField')),
                ('reply_address', models.CharField(blank=True, max_length=255, verbose_name='reply-to email EmailField')),
                ('email_subject', models.CharField(blank=True, max_length=255, verbose_name='email subject CharField')),
                ('create_pdf', models.BooleanField(default=False, verbose_name='create PDF')),
                ('pdf_template', models.CharField(default='django_simple_forms/internal/pdfs/default_pdf.html', max_length=255, verbose_name='PDF template')),
                ('editable_data', models.BooleanField(default=False)),
                ('sort_created', models.BooleanField(default=True, verbose_name='sort by date created (if no table fields specified)')),
                ('form_html', models.CharField(default='django_simple_forms/external/pages/form_page.html', max_length=255, verbose_name='form HTML (external)')),
                ('form_css', models.CharField(default='django_simple_forms/external/css/form_style.css', max_length=255, verbose_name='form CSS (external)')),
                ('success_html', models.CharField(default='django_simple_forms/external/pages/form_submitted.html', max_length=255, verbose_name='success page HTML (external)')),
                ('button_label', models.CharField(default='Submit', max_length=255)),
                ('post_html', models.CharField(default='django_simple_forms/internal/pages/post_view.html', max_length=255, verbose_name='response HTML (internal)')),
                ('post_form_html', models.CharField(default='django_simple_forms/internal/pages/post_form.html', max_length=255, verbose_name='form HTML (internal)')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FormInstance',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('title', models.CharField(max_length=255, unique=True)),
                ('alt_title', models.CharField(blank=True, max_length=255)),
                ('published', models.BooleanField(default=False)),
                ('descriptive_text', ckeditor.fields.RichTextField(blank=True)),
                ('data_json', models.JSONField(blank=True, null=True)),
                ('footer_text', ckeditor.fields.RichTextField(blank=True)),
                ('email_addresses', models.ManyToManyField(blank=True, related_name='form_instances', to='django_simple_forms.EmailAddress')),
                ('form_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='form_instances', to='django_simple_forms.formdefinition')),
            ],
            options={
                'ordering': ['title'],
            },
        ),
        migrations.CreateModel(
            name='FormsSiteProfile',
            fields=[
                ('site_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='sites.site')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('protocol', models.CharField(default='https://', max_length=255)),
            ],
            options={
                'verbose_name': 'site profile',
            },
            bases=('sites.site', models.Model),
            managers=[
                ('objects', django.contrib.sites.models.SiteManager()),
            ],
        ),
        migrations.CreateModel(
            name='InputDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('definition_name', models.CharField(max_length=255, unique=True)),
                ('field_type', models.CharField(choices=[('BooleanField', 'BooleanField'), ('CharField', 'CharField'), ('ChoiceField', 'ChoiceField'), ('DateField', 'DateField'), ('EmailField', 'EmailField'), ('MultipleChoiceField', 'MultipleChoiceField'), ('TimeField', 'TimeField')], default='CharField', max_length=255)),
                ('label', models.CharField(max_length=255)),
                ('label_suffix', models.CharField(default=':', max_length=255)),
                ('min_length', models.PositiveIntegerField(blank=True, null=True, verbose_name='min length (CharField and EmailField only)')),
                ('max_length', models.PositiveIntegerField(blank=True, default=255, null=True, verbose_name='max length (CharField and EmailField only)')),
                ('initial', models.CharField(blank=True, max_length=255, verbose_name='initial value')),
                ('initial_boolean', models.BooleanField(default=False, verbose_name='initial True (BooleanField only)')),
                ('boolean_true', models.CharField(blank=True, max_length=255, verbose_name='boolean true string (if select widget)')),
                ('boolean_false', models.CharField(blank=True, max_length=255, verbose_name='boolean false string (if select widget)')),
                ('empty_value', models.CharField(blank=True, max_length=255, verbose_name='empty value (CharField only)')),
                ('help_text', ckeditor.fields.RichTextField(blank=True)),
                ('strip', models.BooleanField(default=True, verbose_name='strip (CharField only)')),
                ('required', models.BooleanField(default=True)),
                ('widget_type', models.CharField(choices=[('CheckboxInput', 'CheckboxInput'), ('CheckboxSelectMultiple', 'CheckboxSelectMultiple'), ('DateInput', 'DateInput'), ('EmailInput', 'EmailInput'), ('Select', 'Select'), ('SelectMultiple', 'SelectMultiple'), ('Textarea', 'Textarea'), ('TextInput', 'TextInput'), ('TimeInput', 'TimeInput')], default='TextInput', max_length=255)),
                ('widget_autocomplete', models.BooleanField(default=True, verbose_name='autocomplete')),
                ('widget_title', models.CharField(blank=True, max_length=255, verbose_name='title')),
                ('widget_class', models.CharField(blank=True, max_length=255, verbose_name='class')),
            ],
            options={
                'ordering': ['definition_name'],
            },
        ),
        migrations.CreateModel(
            name='InputInstance',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('field_name', models.CharField(max_length=255)),
                ('admin_only', models.BooleanField(default=False)),
                ('sort_order', models.PositiveIntegerField(default=0)),
                ('form_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='input_instances', to='django_simple_forms.formdefinition')),
                ('input_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='input_instances', to='django_simple_forms.inputdefinition')),
            ],
            options={
                'verbose_name': 'input',
                'ordering': ['sort_order'],
            },
        ),
        migrations.CreateModel(
            name='PostCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('category', models.CharField(default='Uncategorized', max_length=255)),
                ('data_json', models.JSONField(blank=True, null=True)),
                ('form_instance', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='post_category', to='django_simple_forms.forminstance')),
            ],
            options={
                'verbose_name': 'form responses',
                'verbose_name_plural': 'form responses',
            },
        ),
        migrations.CreateModel(
            name='PostPDF',
            fields=[
            ],
            options={
                'verbose_name': 'post PDF',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('django_simple_file_handler.privatepdf',),
        ),
        migrations.CreateModel(
            name='TableField',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('column_name', models.CharField(blank=True, max_length=255, verbose_name='custom column header')),
                ('tally_column', models.BooleanField(default=False, verbose_name='tally column (BooleanField only)')),
                ('sort_order', models.PositiveIntegerField(default=0)),
                ('form_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='table_fields', to='django_simple_forms.formdefinition')),
                ('input_instance', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='table_fields', to='django_simple_forms.inputinstance')),
            ],
            options={
                'ordering': ['sort_order'],
            },
        ),
        migrations.CreateModel(
            name='PostInstance',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('data_json', models.JSONField(blank=True, null=True)),
                ('pdf', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='post_instance', to='django_simple_forms.postpdf')),
                ('post_category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='post_instances', to='django_simple_forms.postcategory')),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='ErrorDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('value', models.CharField(max_length=255)),
                ('key', models.CharField(choices=[('required', 'required'), ('invalid', 'invalid (Datefield, EmailField, TimeField only)'), ('invalid_choice', 'invalid_choice (ChoiceField, MultipleChoiceField only)'), ('invalid_list', 'invalid_list (MultipleChoiceField only)'), ('max_length', 'max_length (CharField only)'), ('min_length', 'min_length (CharField only)')], default='required', max_length=255)),
                ('input_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_simple_forms.inputdefinition')),
            ],
            options={
                'verbose_name': 'error message',
            },
        ),
        migrations.CreateModel(
            name='ChoiceDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('value', models.CharField(max_length=255)),
                ('display', models.CharField(max_length=255)),
                ('sort_order', models.PositiveIntegerField(default=0)),
                ('input_definition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_simple_forms.inputdefinition')),
            ],
            options={
                'verbose_name': 'choice',
                'verbose_name_plural': 'choices (ChoiceField and MultipleChoiceField only',
                'ordering': ['sort_order'],
            },
        ),
    ]



from django.core.management.base import (
    BaseCommand,
)


from ...models import (
    EmailAddress,
)


class Command(BaseCommand):
    help = 'Deletes email addresses that do not have any forms assigned'

    def handle(self, *args, **options):
        EmailAddress.objects.filter(
            form_instances=None,
        ).delete()



from collections import (
    OrderedDict,
)
from io import (
    BytesIO,
)
from itertools import (
    chain,
)


from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.contrib.sites.models import (
    Site,
)
from django.contrib.staticfiles.storage import (
    staticfiles_storage,
)
from django.core.exceptions import (
    ObjectDoesNotExist,
)
from django.core.mail import (
    EmailMultiAlternatives,
)
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseRedirect,
)
from django.shortcuts import (
    render,
)
from django.template.loader import (
    get_template,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone,
)
from django.utils.safestring import (
    mark_safe,
)
from django.utils.text import (
    slugify,
)


import xlsxwriter


from .forms import (
    GeneratedForm,
    GeneratedFormPublic,
)
from .models import (
    ChoiceDefinition,
    ErrorDefinition,
    format_data,
    FormInstance,
    FormsSiteProfile,
    InputInstance,
    PostCategory,
    PostInstance,
)


def get_object(app_label, model_name, pk):
    model_class = apps.get_model(
        app_label,
        model_name,
    )
    return model_class.objects.get(
        pk=pk,
    )


def clone_object(obj, clear_attrs=None, attr_name=None, reverse_str=None, reverse_args=None):
    if reverse_args is None:
        reverse_args = []
    obj.pk = None
    obj.id = None
    if clear_attrs:
        for attr in clear_attrs:
            setattr(obj, attr, None)
    if attr_name:
        clone_num = 1
        found_copy = True
        clone_name = ''
        while found_copy:
            clone_name = f'{getattr(obj, attr_name)} copy {str(clone_num)}'
            try:
                obj.__class__.objects.get(
                    **{attr_name: clone_name},
                )
                clone_num += 1
            except ObjectDoesNotExist:
                found_copy = False
        setattr(obj, attr_name, clone_name)
    obj.save()
    if not reverse_str:
        reverse_str = f'admin:{obj._meta.app_label}_{obj._meta.model_name}_change'
    reverse_args += [obj.pk]
    redirect = HttpResponseRedirect(
        reverse(
            reverse_str,
            args=reverse_args,
        )
    )
    return obj, redirect


def clone_related(obj, related, attr):
    for item in related:
        item.pk = None
        item.id = None
        setattr(item, attr, obj)
        item.save()


def clone_input_definition(request, app_label, model_name, pk):
    obj = get_object(
        app_label,
        model_name,
        pk,
    )
    error_defs = ErrorDefinition.objects.filter(
        input_definition=obj,
    )
    choice_defs = ChoiceDefinition.objects.filter(
        input_definition=obj,
    )
    related = chain(
        error_defs,
        choice_defs,
    )
    clone = clone_object(
        obj,
        attr_name='definition_name',
    )
    clone_related(
        clone[0],
        related,
        'input_definition',
    )
    return clone[1]


def clone_email_address(request, app_label, model_name, pk):
    obj = get_object(
        app_label,
        model_name,
        pk,
    )
    form_instances = obj.form_instances.all()
    clone = clone_object(
        obj,
        attr_name='recipient_name',
    )
    clone[0].email = 'example@example.com'
    clone[0].save()
    for item in form_instances:
        item.email_addresses.add(clone[0])
    return clone[1]


def clone_form_definition(request, app_label, model_name, pk):
    obj = get_object(
        app_label,
        model_name,
        pk,
    )
    related = InputInstance.objects.filter(
        form_definition=obj,
    )
    clone = clone_object(
        obj,
        attr_name='definition_name',
    )
    clone_related(
        clone[0],
        related,
        'form_definition',
    )
    return clone[1]


def clone_form_instance(request, app_label, model_name, pk):
    obj = get_object(
        app_label,
        model_name,
        pk,
    )
    email_addresses = obj.email_addresses.all()
    clone = clone_object(
        obj,
        attr_name='title',
    )
    clone[0].published = False
    clone[0].save()
    for item in email_addresses:
        clone[0].email_addresses.add(item)
    return clone[1]


def clone_post_instance(request, app_label, model_name, pk):
    obj = get_object(
        'django_simple_forms',
        'postinstance',
        pk,
    )
    clone = clone_object(
        obj,
        clear_attrs=[
            'pdf',
        ],
        reverse_str='django_simple_forms:form_page_admin',
        reverse_args=[
            app_label,
            model_name,
        ],
    )
    return clone[1]


def form_page(request, pk, title_slug):
    try:
        form_instance = FormInstance.objects.get(pk=pk)
        if not form_instance.published:
            raise Http404()
    except (ValueError, ObjectDoesNotExist) as error:
        raise Http404()
    ''' Begin getting time elements '''
    form_submit_time = timezone.now().timestamp()
    if 'form_load_time' in request.session:
        form_load_time = request.session['form_load_time']
        del request.session['form_load_time']
    else:
        form_load_time = form_submit_time
    ''' End getting time elements '''
    form_definition = form_instance.form_definition
    page_template = form_definition.form_html
    success_template = form_definition.success_html
    dict_data = form_instance.dict_data()
    for field in dict_data['field_dicts']:
        if field['admin_only']:
            field['widget_data']['widget_type'] = 'HiddenInput'
            field['field_data']['field_args']['required'] = False
    form = GeneratedFormPublic(
        dict_data=dict_data,
    )
    if request.method == 'POST':
        form = GeneratedFormPublic(
            request.POST,
            dict_data=dict_data,
        )
        if form.is_valid():
            cleaned_data = form.cleaned_data
            context = {
                'title': form_instance.title,
                'css_style': staticfiles_storage.url(form_definition.form_css),
            }
            ''' Begin checking how quickly form was submitted '''
            if form_submit_time - form_load_time > 1:
                ''' End checking how quickly form was submitted '''
                ''' Begin checking for input in hidden field '''
                if not cleaned_data['email']:
                    ''' End checking for input in hidden field '''
                    ''' Note: Optional reCAPTCHA is verified in the page'''
                    ''' Begin removing security data '''
                    cleaned_data.pop('email')
                    try:
                        cleaned_data.pop('captcha')
                    except KeyError:
                        pass
                    ''' End removing security data '''
                    post_instance = PostInstance.objects.create_instance(
                        form_instance=form_instance,
                        cleaned_data=cleaned_data,
                    )
                    if form_definition.send_email:
                        try:
                            from_address = settings.FORMS_FROM_ADDRESS
                        except AttributeError:
                            from_address = settings.DEFAULT_FROM_EMAIL
                        input_instances = form_definition.input_instances
                        first_instance = input_instances.all().filter(
                            field_name=form_definition.reply_first,
                        ).first()
                        last_instance = input_instances.all().filter(
                            field_name=form_definition.reply_last,
                        ).first()
                        address_instance = input_instances.all().filter(
                            field_name=form_definition.reply_address,
                        ).first()
                        if address_instance and address_instance.input_definition.field_type == 'EmailField':
                            reply_address = cleaned_data[address_instance.field_name]
                            reply_to = reply_address
                            if first_instance and first_instance.input_definition.field_type == 'CharField':
                                reply_first = cleaned_data[first_instance.field_name]
                                reply_to = f'"{reply_first}" <{reply_address}>'
                                if last_instance and last_instance.input_definition.field_type == 'CharField':
                                    reply_last = cleaned_data[last_instance.field_name]
                                    reply_to = f'"{reply_first} {reply_last}" <{reply_address}>'
                        else:
                            try:
                                reply_to = settings.FORMS_REPLY_ADDRESS
                            except AttributeError:
                                reply_to = from_address
                        subject_instance = input_instances.all().filter(
                            field_name=form_definition.email_subject,
                        ).first()
                        prefix_str = post_instance.post_category.form_instance.form_definition.subject_prefix
                        if subject_instance and subject_instance.input_definition.field_type == 'CharField':
                            field_str = cleaned_data[subject_instance.field_name]
                        else:
                            field_str = ''
                        if prefix_str and field_str:
                            space_str = ' '
                        else:
                            space_str = ''
                        subject_str = f'{prefix_str}{space_str}{field_str}'
                        text_template = form_definition.email_txt
                        html_template = form_definition.email_html
                        subject = subject_str
                        site_domain = Site.objects.get_current().domain
                        site_profile = FormsSiteProfile.objects.filter(
                            domain=site_domain,
                        ).first()
                        if post_instance.pdf:
                            pdf_url = f'{site_profile.protocol_domain()}{post_instance.pdf.proxy_url()}'
                        else:
                            pdf_url = ''
                        email_content = {
                            'subject': subject,
                            'form_name': form_instance.title,
                            'output_dict': post_instance.output_dict(),
                            'output_no_br': post_instance.output_no_br(),
                            'output_formatted_date_time': post_instance.output_formatted_date_time(),
                            'output_text': post_instance.output_text(),
                            'pdf_url': pdf_url,
                        }
                        email_txt = get_template(text_template).render(email_content)
                        email_html = get_template(html_template).render(email_content)
                        recipients = [
                            f'"{address.recipient_name}" <{address.email}>'
                            for address in form_instance.email_addresses.all()
                        ]
                        message = EmailMultiAlternatives(
                            subject,
                            email_txt,
                            from_address,
                            recipients,
                            reply_to=[reply_to],
                        )
                        message.attach_alternative(email_html, 'text/html')
                        message.send(fail_silently=False)
            return render(
                request,
                success_template,
                context,
            )
    ''' Begin setting when form was loaded '''
    request.session['form_load_time'] = timezone.now().timestamp()
    ''' End setting when form was loaded '''
    context = {
        'title': form_instance.title,
        'descriptive_text': mark_safe(form_instance.descriptive_text),
        'footer_text': mark_safe(form_instance.footer_text),
        'button_label': form_instance.form_definition.button_label,
        'css_style': staticfiles_storage.url(form_instance.form_definition.form_css),
        'form': form,
    }
    return render(
        request,
        page_template,
        context,
    )


def form_page_admin(request, app_label, model_name, pk, add_instance=False):
    post_instance = PostInstance.objects.none()
    if add_instance:
        if not request.user.has_perm('django_simple_forms.add_postinstance'):
            raise Http404()
        try:
            post_category = PostCategory.objects.get(pk=pk)
        except (ValueError, ObjectDoesNotExist) as error:
            raise Http404()
        form_instance = post_category.form_instance
        if not form_instance:
            raise Http404()
        title = f'Add response: {form_instance.title}'
        dict_data = form_instance.dict_data()
    else:
        post_instance = PostInstance.objects.get(pk=pk)
        post_category = post_instance.post_category
        form_instance = post_category.form_instance
        if not form_instance:
            raise Http404()
        if not form_instance.form_definition.editable_data\
                or not request.user.has_perm('django_simple_forms.change_postinstance'):
            raise Http404()
        title = f'Edit response: {form_instance.title}'
        dict_data = form_instance.dict_data()
        for key, value in post_instance.output_no_br().items():
            for field in dict_data['field_dicts']:
                if field['field_data']['field_name'] == key:
                    try:
                        if field['field_data']['field_type'] == 'MultipleChoiceField':
                            field['field_data']['field_args']['initial'] = value[1:]
                        else:
                            field['field_data']['field_args']['initial'] = value[1]
                        if field['field_data']['field_type'] == 'TypedChoiceField':
                            if field['field_data']['field_args']['initial']:
                                field['field_data']['field_args']['initial'] = 'true'
                            else:
                                field['field_data']['field_args']['initial'] = 'false'
                    except IndexError:
                        pass
    page_template = form_instance.form_definition.post_form_html
    reverse_str = f'admin:{app_label}_{model_name}_change'
    form = GeneratedForm(
        dict_data=dict_data,
    )
    if request.method == 'POST':
        form = GeneratedForm(
            request.POST,
            dict_data=dict_data,
        )
        category_return = HttpResponseRedirect(
            reverse(
                reverse_str,
                args=(post_category.pk,),
            )
        )
        if 'cancel' in request.POST:
            return category_return
        if form.is_valid():
            cleaned_data = form.cleaned_data
            if add_instance:
                PostInstance.objects.create_instance(
                    form_instance=form_instance,
                    cleaned_data=cleaned_data,
                )
            else:
                stored_dict = format_data(
                    form_instance=form_instance,
                    cleaned_data=cleaned_data,
                )
                if post_instance:
                    post_instance.data_json = stored_dict
                    post_instance.save()
            return category_return
    context = {
        'title': title,
        'form': form,
    }
    return render(
        request,
        page_template,
        context,
    )


def form_page_admin_add(request, app_label, model_name, pk):
    return form_page_admin(
        request,
        app_label,
        model_name,
        pk,
        add_instance=True,
    )


def post_view(request, app_label, model_name, pk):
    try:
        post_instance = PostInstance.objects.get(pk=pk)
    except (ValueError, ObjectDoesNotExist) as error:
        raise Http404()
    if not request.user.has_perm('django_simple_forms.view_postinstance'):
        raise Http404()
    title = f'View response: {post_instance.post_category.category}'
    delete_post = False
    delete_perm = False
    if request.user.has_perm('django_simple_forms.delete_postinstance'):
        delete_perm = True
    if request.method == 'POST':
        reverse_str = f'admin:{app_label}_{model_name}_change'
        category_return = HttpResponseRedirect(
            reverse(
                reverse_str,
                args=(post_instance.post_category.pk,),
            )
        )
        if 'return' in request.POST:
            return category_return
        if 'delete' in request.POST:
            delete_post = True
        if 'confirm' in request.POST:
            post_instance.delete()
            return category_return
    context = {
        'title': title,
        'output_dict': post_instance.output_dict(),
        'output_formatted_date_time': post_instance.output_formatted_date_time(),
        'created': post_instance.created,
        'updated': post_instance.updated,
        'delete_perm': delete_perm,
        'delete_post': delete_post,
    }
    if post_instance.post_category.form_instance:
        page_template = post_instance.post_category.form_instance.form_definition.post_html
    else:
        page_template = 'django_simple_forms/internal/pages/post_view.html'
    return render(
        request,
        page_template,
        context,
    )


def post_spreadsheet(request, app_label, model_name, pk):
    category = get_object(
        app_label,
        model_name,
        pk,
    )
    post_instances = category.post_instances
    form_slug = slugify(category.category)
    datetime_str = str(timezone.now()).split('.')[0].replace(' ', '_').replace(':', '.')
    filename = f'{form_slug}_{datetime_str}.xlsx'
    output = BytesIO()
    workbook = xlsxwriter.Workbook(output)
    format_title = workbook.add_format(
        {
            'bold': True,
            'font_size': 16,
        }
    )
    format_header = workbook.add_format(
        {
            'bold': True,
            'font_size': 12,
        }
    )
    format_header.set_bottom(1)
    format_header.set_text_wrap()
    format_cell_data = workbook.add_format(
        {
            'align': 'top',
            'font_size': 12,
        }
    )
    format_cell_data.set_text_wrap()
    format_tallies = workbook.add_format(
        {
            'bold': True,
            'font_size': 12,
        }
    )
    format_tallies.set_top(1)
    format_tallies.set_text_wrap()
    worksheet = workbook.add_worksheet()
    worksheet.write(0, 0, category.category, format_title)
    col_dict = OrderedDict()
    for post in post_instances.all():
        for key, value in post.output_no_br().items():
            if key not in col_dict:
                col_dict[key] = value[0]
    col_keys = []
    header_row = 2
    header_col = 0
    for key, value in col_dict.items():
        col_keys.append(key)
        worksheet.write(header_row, header_col, value, format_header)
        header_col += 1
    last_col = len(col_dict) - 1
    worksheet.set_column(0, last_col, 20)
    data_row = 3
    for post in post_instances.all():                           # Doing this iteration a second time
        for key, value in post.output_no_br().items():
            data_col = col_keys.index(key)
            if len(value) > 2:
                value_list = value[1:]
                cell_data = ''
                for item in value_list:
                    cell_data += f'{str(item)},\r\n'
                cell_data = cell_data[:-3]
            else:
                try:
                    if value[1] is True:
                        cell_data = '✓'
                    elif value[1] is False:
                        cell_data = '✕'
                    else:
                        cell_data = value[1]
                except IndexError:
                    cell_data = ''
            worksheet.write(data_row, data_col, cell_data, format_cell_data)
        data_row += 1
    tallies = ['' for key in col_keys]
    for field_name in category.data_json['tally_fields']:
        total = 0
        for instance in category.post_instances.all():
            for item in instance.data_json['data_dicts']:
                for key, value in item.items():
                    if key == field_name and value[1]:
                        total += 1
        tally_str = f'Total:\r\n{str(total)}'
        tallies[col_keys.index(field_name)] = str(tally_str)
    for index in range(len(tallies)):
        worksheet.write(data_row, index, tallies[index], format_tallies)
    workbook.close()
    response = HttpResponse(
        content_type='application/vnd.ms-excel',
    )
    response['Content-Disposition'] = f'attachment;filename="{filename}"'
    response.write(output.getvalue())
    return response



from functools import (
    partial,
)
from operator import (
    itemgetter,
)


from django import (
    forms,
)
from django.core.exceptions import (
    ObjectDoesNotExist,
)
from django.forms import (
    ValidationError,
)
from django.contrib import (
    admin
)
from django.contrib.staticfiles.storage import (
    staticfiles_storage,
)
from django.http import (
    HttpResponseRedirect,
)
from django.urls import (
    reverse,
)
from django.utils import (
    timezone,
)
from django.utils.formats import (
    localize,
)
from django.utils.safestring import (
    mark_safe,
)


from adminsortable2.admin import (
    SortableInlineAdminMixin,
)


from .models import (
    ChoiceDefinition,
    EmailAddress,
    ErrorDefinition,
    FormDefinition,
    FormInstance,
    FormsSiteProfile,
    InputDefinition,
    InputInstance,
    PostCategory,
    TableField,
)


class BaseAdmin(admin.ModelAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_css = 'admin/css/django_simple_forms.css'
    actions = None
    readonly_fields = [
        'created',
        'updated',
    ]
    bottom_fieldsets = [
        (
            'Date and time information', {
                'fields': [
                    'created',
                    'updated',
                ],
                'classes': [
                    'collapse',
                ],
            }
        ),
    ]
    fieldsets = bottom_fieldsets
    list_per_page = 20

    @property
    def media(self):
        css = {
            'all': (
                self.base_css,
            ),
        }
        return forms.Media(css=css)


class FormsSiteProfileAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'protocol',
                        'domain',
                        'name',
                    ]
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    search_fields = [
        'protocol',
        'domain',
        'name',
    ]
    list_display = [
        'name',
        'domain',
    ]
    ordering = [
        'name',
    ]


admin.site.register(
    FormsSiteProfile,
    FormsSiteProfileAdmin,
)


class CloneableAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'clone_link',
        ] + self.readonly_fields
        self.bottom_fieldsets = [
            (
                'Clone', {
                    'fields': [
                        'clone_link',
                    ],
                }
            ),
        ] + self.bottom_fieldsets
        self.list_display = [
            'clone_link',
        ]
        self.view_name = ''

    def clone_link(self, obj):
        view_str = f'django_simple_forms:{self.view_name}'
        if obj.pk:
            url_str = reverse(
                view_str,
                kwargs={
                    'app_label': obj._meta.app_label,
                    'model_name': obj._meta.model_name,
                    'pk': obj.pk,
                },
            )
            link_str = f'<a href="{url_str}"><input type="button" class= "small_admin_button" value="Clone"></a>'
        else:
            link_str = 'Click "save and continue editing" to enable cloning.'
        return mark_safe(link_str)

    clone_link.short_description = 'clone'


class FormInstanceInline(admin.TabularInline):
    model = FormInstance.email_addresses.through
    verbose_name = 'associated form'
    verbose_name_plural = 'associated forms'
    extra = 0


class EmailAddressAdmin(CloneableAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'inst_count',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'recipient_name',
                        'email',
                    ],
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets
        self.list_display = [
            'recipient_name',
            'email',
            'inst_count',
        ] + self.list_display
        self.view_name = 'clone_email_address'

    inlines = [
        FormInstanceInline,
    ]
    search_fields = [
        'recipient_name',
        'email',
    ]
    ordering = [
        'recipient_name',
    ]


admin.site.register(
    EmailAddress,
    EmailAddressAdmin,
)


class ErrorDefinitionInline(admin.TabularInline):
    model = ErrorDefinition
    fields = [
        'key',
        'value',
    ]
    extra = 0


class ChoiceDefinitionInline(SortableInlineAdminMixin, admin.TabularInline):
    model = ChoiceDefinition
    exclude = [
    ]
    extra = 0


class InputDefinitionAdmin(CloneableAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'def_links',
            'inst_count',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                'Field', {
                    'fields': [
                        'definition_name',
                        'field_type',
                        'label',
                        'label_suffix',
                    ],
                }
            ),
            (
                'Field arguments', {
                    'fields': [
                        'min_length',
                        'max_length',
                        'initial',
                        'initial_boolean',
                        'boolean_true',
                        'boolean_false',
                        'empty_value',
                        'help_text',
                        'strip',
                        'required',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
            (
                'Widget', {
                    'fields': [
                        'widget_type',
                    ]
                }
            ),
            (
                'Widget attributes', {
                    'fields': [
                        'widget_autocomplete',
                        'widget_title',
                        'widget_class',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
            (
                'Form definitions', {
                    'fields': [
                        'def_links',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets
        self.list_display = [
            'definition_name',
            'field_type',
            'widget_type',
            'required',
            'inst_count',
        ] + self.list_display
        self.view_name = 'clone_input_definition'

    inlines = [
        ErrorDefinitionInline,
        ChoiceDefinitionInline,
    ]
    search_fields = [
        'definition_name',
    ]
    ordering = [
        'definition_name',
    ]

    def def_links(self, obj):
        link_str = ''
        for item in obj.input_instances.all():
            url_str = reverse(
                f'admin:{item.form_definition._meta.app_label}_{item.form_definition._meta.model_name}_change',
                args=(item.form_definition.pk,),
            )
            link_str = f'{link_str}<a href="{url_str}">{item.form_definition.definition_name}</a><br><br>'
        if link_str:
            link_str = link_str[:-8]
        else:
            link_str = 'Input is not being used in any form definitions.'
        return mark_safe(link_str)

    def_links.short_description = 'form definitions'


admin.site.register(
    InputDefinition,
    InputDefinitionAdmin,
)


class InputInstanceInlineForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        field_name = cleaned_data.get('field_name')
        if field_name == 'email':
            raise ValidationError(
                'The field name "email" is reserved for the honeypot field.',
            )

    class Meta:
        model = InputInstance
        exclude = [
        ]


class InputInstanceInline(SortableInlineAdminMixin, admin.TabularInline):
    form = InputInstanceInlineForm
    model = InputInstance
    extra = 0


class TableFieldInline(SortableInlineAdminMixin, admin.TabularInline):
    def get_formset(self, request, obj=None, **kwargs):
        kwargs['formfield_callback'] = partial(
            self.formfield_for_dbfield,
            request=request,
            obj=obj,
        )
        return super().get_formset(request, obj, **kwargs)

    def formfield_for_dbfield(self, db_field, **kwargs):
        form_definition = kwargs.pop('obj', None)
        formfield = super().formfield_for_dbfield(db_field, **kwargs)
        if db_field.name == 'input_instance' and form_definition:
            formfield.queryset = form_definition.input_instances
        return formfield

    model = TableField
    extra = 0


class FormDefinitionAdmin(CloneableAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = None
        self.readonly_fields = [
            'inst_links',
            'inst_count',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'definition_name',
                        'editable_data',
                        'sort_created',
                    ],
                }
            ),
            (
                'Email settings', {
                    'fields': [
                        'send_email',
                        'email_html',
                        'email_txt',
                        'reply_first',
                        'reply_last',
                        'reply_address',
                        'subject_prefix',
                        'email_subject',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
            (
                'PDF settings', {
                    'fields': [
                        'create_pdf',
                        'pdf_template',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
            (
                'Page settings', {
                    'fields': [
                        'form_html',
                        'form_css',
                        'success_html',
                        'button_label',
                        'post_html',
                        'post_form_html',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
            (
                'Form instances', {
                    'fields': [
                        'inst_links',
                    ],
                    'classes': [
                        'collapse',
                    ],
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets
        self.list_display = [
            'definition_name',
            'inst_count',
        ] + self.list_display
        self.view_name = 'clone_form_definition'

    inlines = [
        InputInstanceInline,
        TableFieldInline,
    ]
    search_fields = [
        'definition_name',
    ]
    ordering = [
        'definition_name',
    ]

    def inst_links(self, obj):
        if obj.pk:
            link_str = ''
            for item in obj.form_instances.all():
                url_str = reverse(
                    f'admin:{item._meta.app_label}_{item._meta.model_name}_change',
                    args=(item.pk,),
                )
                link_str = f'{link_str}<a href="{url_str}">{item.title}</a><br><br>'
            if link_str:
                link_str = link_str[:-8]
        else:
            link_str = 'Click "save and continue editing" to get a link.'
        return mark_safe(link_str)

    inst_links.short_description = 'form instances'

    def get_form(self, request, obj=None, **kwargs):
        self.obj = obj
        return super().get_form(request, obj, **kwargs)

    def save_model(self, request, obj, form, change):
        obj.from_admin = True
        super().save_model(request, obj, form, change)

    def save_related(self, request, form, formsets, change):
        form.save_m2m()
        for formset in formsets:
            self.save_formset(request, form, formset, change=change)
        try:
            self.obj.downstream_update()
        except AttributeError:
            pass


admin.site.register(
    FormDefinition,
    FormDefinitionAdmin,
)


class FormInstanceAdminForm(forms.ModelForm):
    email_addresses = forms.ModelMultipleChoiceField(
        queryset=EmailAddress.objects.order_by(
            'email',
        ),
        label='Email addresses',
        required=False,
        widget=admin.widgets.FilteredSelectMultiple(
            'email addresses',
            False,
        )
    )

    class Meta:
        model = FormInstance
        exclude = [
        ]


class FormInstanceAdmin(CloneableAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            'form_link',
            'def_link',
            'cat_link',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'title',
                        'alt_title',
                        'form_definition',
                        'published',
                        'descriptive_text',
                        'footer_text',
                        'email_addresses',
                    ],
                }
            ),
        ]
        self.bottom_fieldsets[0][1]['fields'] = self.bottom_fieldsets[0][1]['fields'] + [
            'form_link',
            'def_link',
            'cat_link',
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets
        self.list_display = [
            'title',
            'published',
            'form_link',
        ] + self.list_display
        self.view_name = 'clone_form_instance'

    def get_form(self, request, obj=None, **kwargs):
        if obj and not self.has_change_permission(request, obj):
            return super().get_form(request, obj, **kwargs)
        return FormInstanceAdminForm

    search_fields = [
        'title',
    ]

    def form_link(self, obj):
        if obj.published:
            link_str = f'<a href="{obj.form_url()}" target="_blank">Form page</a>'
        else:
            link_str = 'Unpublished'
        return mark_safe(link_str)

    form_link.short_description = 'page'

    def def_link(self, obj):
        if obj.pk:
            url_str = reverse(
                f'admin:{obj.form_definition._meta.app_label}_{obj.form_definition._meta.model_name}_change',
                args=(obj.form_definition.pk,),
            )
            link_str = f'<a href="{url_str}">{obj.form_definition.definition_name}</a>'
        else:
            link_str = 'Click "save and continue editing" to get a link.'
        return mark_safe(link_str)

    def_link.short_description = 'form definition'

    def cat_link(self, obj):
        link_str = 'Click "save and continue editing" to get a link.'
        if obj.pk:
            try:
                url_str = reverse(
                    f'admin:{obj.post_category._meta.app_label}_{obj.post_category._meta.model_name}_change',
                    args=(obj.post_category.pk,),
                )
                link_str = f'<a href="{url_str}">Form responses</a>'
            except ObjectDoesNotExist:
                pass
        return mark_safe(link_str)

    cat_link.short_description = 'responses'


admin.site.register(
    FormInstance,
    FormInstanceAdmin,
)


class PostCategoryAdmin(BaseAdmin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = None
        self.obj = None
        self.obj_app_label = None
        self.obj_model_name = None
        self.obj_pk = None
        self.obj_form_instance = None
        self.readonly_fields = [
           'add_response',
           'spreadsheet_export',
           'responses',
           'inst_link',
           'inst_count',
        ] + self.readonly_fields
        self.top_fieldsets = [
            (
                None, {
                    'fields': [
                        'inst_link',
                        'add_response',
                        'inst_count',
                        'spreadsheet_export',
                        'responses',
                    ],
                }
            ),
        ]
        self.fieldsets = self.top_fieldsets + self.bottom_fieldsets

    search_fields = [
        'category',
    ]
    list_display = [
        'form_title',
        'inst_count',
        'form_instance_exists',
    ]
    ordering = [
        'category',
    ]

    def form_instance_exists(self, obj):
        if obj.form_instance:
            image_url = 'admin/img/icon-yes.svg'
            alt = 'True'
        else:
            image_url = 'admin/img/icon-no.svg'
            alt = 'False'
        image_str = f'<img src="{staticfiles_storage.url(image_url)}" alt="{alt}">'
        return mark_safe(image_str)

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def get_form(self, request, obj=None, **kwargs):
        self.request = request
        self.obj = obj
        return super().get_form(request, obj, **kwargs)

    def form_title(self, obj):
        return obj.category

    def inst_link(self, obj):
        if obj.form_instance:
            page_url = obj.form_instance.form_url()
            admin_url = reverse(
                f'admin:{obj.form_instance._meta.app_label}_{obj.form_instance._meta.model_name}_change',
                args=(obj.form_instance.pk,),
            )
            if obj.form_instance.published:
                page_link = f'<a href="{page_url}" target="_blank">Form page</a>'
            else:
                page_link = 'Form not published'
            link_str = f'{page_link}' \
                       f'<br>' \
                       f'<br>' \
                       f'<a href="{admin_url}">Form instance admin</a>'
        else:
            link_str = 'Form instance does not exist.'
        return mark_safe(link_str)

    inst_link.short_description = 'form instance'

    def add_response(self, obj):
        url_str = reverse(
            'django_simple_forms:form_page_admin_add',
            kwargs={
                'app_label': obj._meta.app_label,
                'model_name': obj._meta.model_name,
                'pk': obj.pk,
            },
        )
        if not obj.form_instance:
            return 'Form instance does not exist.'
        if self.request.user.has_perm('django_simple_forms.add_postinstance'):
            return mark_safe(
                f'<a href="{url_str}">'
                f'<input type="button" class= "small_admin_button" value="Add response">'
                f'</a>'
            )
        else:
            return ''

    def spreadsheet_export(self, obj):
        url_str = reverse(
            'django_simple_forms:post_spreadsheet',
            kwargs={
                'app_label': obj._meta.app_label,
                'model_name': obj._meta.model_name,
                'pk': obj.pk,
            },
        )
        return mark_safe(
            f'<a href="{url_str}">'
            f'<input type="button" class= "small_admin_button" value="Download responses">'
            f'</a>'
        )

    def trim_list(self, post, name):
        try:
            trimmed_list = post.output_formatted_date_time()[name][1:]
        except (IndexError, KeyError) as error:
            trimmed_list = []
        return trimmed_list

    def responses(self, obj):
        field_names = [item for field in obj.data_json['table_fields'] for item in field]
        headers_list = ['Created:']
        if obj.data_json['table_fields']:
            headers_list += [
                value for header_dict in obj.data_json['table_fields'] for key, value in header_dict.items()
            ]
        if not obj.data_json['sort_created'] and obj.data_json['table_fields']:
            headers_list.pop(0)
        data_lists = [
            [[post.created] for x in range(len(headers_list) - len(field_names))]
            + [self.trim_list(post, name) for name in field_names]
            + [post.pk]
            for post in obj.post_instances.all()
        ]
        for index in reversed(range(len(headers_list) + 1)):
            data_lists = sorted(data_lists, key=itemgetter(index))
        if obj.data_json['sort_created'] or not obj.data_json['table_fields']:
            for unformatted_row in data_lists:
                unformatted_row[0][0] = localize(timezone.localtime(unformatted_row[0][0]))
        headers_list += [
            'Data:',
            'PDF:',
            'Clone:',
        ]
        html_str = '<div id="responses_container">' \
                   '<table id="responses_table">'
        html_str += f'<tr>'
        for header in headers_list:
            html_str += f'<th><strong>{header}</strong></th>'
        html_str += f'</tr>'
        for post_list in data_lists:
            row_list = post_list[:-1]
            post = obj.post_instances.all().filter(
                pk=post_list[-1]
            ).first()
            view_link_str = '—'
            if self.request.user.has_perm('django_simple_forms.view_postinstance'):
                url_str_view = reverse(
                    'django_simple_forms:post_view',
                    kwargs={
                        'app_label': obj._meta.app_label,
                        'model_name': obj._meta.model_name,
                        'pk': post.pk,
                    },
                )
                view_link_str = f'<a href="{url_str_view}">View</a>'
            edit_link_str = '—'
            if post.post_category.form_instance:
                if post.post_category.form_instance.form_definition.editable_data\
                        and self.request.user.has_perm('django_simple_forms.change_postinstance'):
                    url_str_edit = reverse(
                        'django_simple_forms:form_page_admin',
                        kwargs={
                            'app_label': obj._meta.app_label,
                            'model_name': obj._meta.model_name,
                            'pk': post.pk,
                        },
                    )
                    edit_link_str = f'<a href="{url_str_edit}">Edit</a>'
            data_link_str = f'{view_link_str} | {edit_link_str}'
            row_list.append([data_link_str])
            pdf_link_str = '—'
            if post.pdf and self.request.user.has_perm('django_simple_forms.view_postpdf'):
                pdf_link_str = f'<a href=" {post.pdf.proxy_url()}" target="_blank">View</a>'
            row_list.append([pdf_link_str])
            clone_button_str = '—'
            if post.post_category.form_instance:
                if post.post_category.form_instance.form_definition.editable_data\
                        and self.request.user.has_perm('django_simple_forms.add_postinstance')\
                        and self.request.user.has_perm('django_simple_forms.change_postinstance'):
                    url_str_clone = reverse(
                        'django_simple_forms:clone_post_instance',
                        kwargs={
                            'app_label': obj._meta.app_label,
                            'model_name': obj._meta.model_name,
                            'pk': post.pk,
                        },
                    )
                    clone_button_str = f'<a href="{url_str_clone}"><input type="button" class= "small_admin_button" value="Clone"></a>'
            row_list.append([clone_button_str])
            html_str += '<tr>'
            for cell_list in row_list:
                if cell_list == row_list[-1]:
                    td_class = 'post_instance no_right_border'
                else:
                    td_class = 'post_instance'
                cell_str = ''
                for data_item in cell_list:
                    if isinstance(data_item, bool):
                        image_url = 'admin/img/icon-unknown.svg'
                        alt = 'Null'
                        if data_item:
                            image_url = 'admin/img/icon-yes.svg'
                            alt = 'True'
                        if not data_item:
                            image_url = 'admin/img/icon-no.svg'
                            alt = 'False'
                        cell_str = f'<img src="{staticfiles_storage.url(image_url)}" alt="{alt}">  '
                    else:
                        cell_str += f'{str(data_item)}, '
                if cell_str:
                    cell_str = cell_str[:-2]
                html_str += f'<td class="{td_class}">{cell_str}</td>'
            html_str += '</tr>'
        col_keys = []
        if obj.data_json:
            col_keys = [key for header_dict in obj.data_json['table_fields'] for key, value in header_dict.items()]
        tally_fields = obj.data_json['tally_fields']
        tallies = []
        for field_name in col_keys:
            if field_name in tally_fields:
                total = 0
                for instance in obj.post_instances.all():
                    for item in instance.data_json['data_dicts']:
                        for key, value in item.items():
                            if key == field_name and value[1]:
                                total += 1
                tally = str(total)
            else:
                tally = ''
            tallies.append(tally)
        tallies += ['', '', '']
        tally_row = False
        tally_html = '<tr class="tally_row">'
        for tally_str in tallies:
            if tally_str:
                tally_row = True
                tally_str = f'<u>Total</u>:<br>{tally_str}'
            tally_html += f'<td class="post_instance no_right_border tally_cell">{tally_str}</td>'
        tally_html += '</tr>'
        if tally_row:
            html_str += tally_html
        html_str += '</table>' \
                    '</div>'
        return mark_safe(html_str)

    @property
    def media(self):
        css = {
            'all': (
                self.base_css,
            ),
        }
        if self.obj:
            if self.obj.form_instance:
                css = {
                    'all': (
                        self.base_css,
                        'admin/css/delete_button.css',
                    ),
                }
        return forms.Media(css=css)

    def delete_model(self, request, obj):
        self.obj_app_label = obj._meta.app_label
        self.obj_model_name = obj._meta.model_name
        self.obj_pk = obj.pk
        self.obj_form_instance = obj.form_instance
        obj.delete()

    def response_delete(self, request, obj_display, obj_id):
        if self.obj_form_instance:
            reverse_str = f'admin:{self.obj_app_label}_{self.obj_model_name}_change'
            reverse_args = [self.obj_pk]
            redirect = HttpResponseRedirect(
                reverse(
                    reverse_str,
                    args=reverse_args,
                )
            )
        else:
            reverse_str = f'admin:{self.obj_app_label}_{self.obj_model_name}_changelist'
            redirect = HttpResponseRedirect(
                reverse(
                    reverse_str,
                )
            )
        return redirect


admin.site.register(
    PostCategory,
    PostCategoryAdmin,
)



from django.conf import (
    settings,
)
from django.forms import (
    BooleanField,
    CharField,
    CheckboxInput,
    CheckboxSelectMultiple,
    ChoiceField,
    DateField,
    DateInput,
    EmailField,
    EmailInput,
    Form,
    HiddenInput,
    MultipleChoiceField,
    Select,
    SelectMultiple,
    Textarea,
    TextInput,
    TimeField,
    TimeInput,
    TypedChoiceField,
)


def get_recaptcha_widget():
    from captcha.widgets import (
        ReCaptchaV2Checkbox,
        ReCaptchaV2Invisible,
        ReCaptchaV3,
    )
    try:
        captcha_attrs = settings.FORMS_RECAPTCHA_ATTRS
    except AttributeError:
        captcha_attrs = {}
    try:
        captcha_params = settings.FORMS_RECAPTCHA_PARAMS
    except AttributeError:
        captcha_params = {}
    widget_dict = {
        1: ReCaptchaV2Checkbox(
            attrs=captcha_attrs,
            api_params=captcha_params,
        ),
        2: ReCaptchaV2Invisible(
            attrs=captcha_attrs,
            api_params=captcha_params,
        ),
        3: ReCaptchaV3(
            attrs=captcha_attrs,
            api_params=captcha_params,
        ),
    }
    try:
        widget_type = settings.FORMS_RECAPTCHA_TYPE
    except AttributeError:
        widget_type = 1
    return widget_dict[widget_type]


class GeneratedForm(Form):
    def __init__(self, *args, **kwargs):
        dict_data = kwargs.pop('dict_data')
        super().__init__(*args, **kwargs)
        field_dicts = dict_data['field_dicts']
        for item in field_dicts:
            field_data = item['field_data']
            added_field = eval(field_data['field_type'])()
            widget_data = item['widget_data']
            added_widget = eval(widget_data['widget_type'])()
            if widget_data['widget_attrs']:
                added_widget.attrs = widget_data['widget_attrs']
            added_field.widget = added_widget
            try:
                choice_dicts = field_data['field_args'].pop('choice_dicts')
                choice_tuples = []
                for choice in choice_dicts:
                    for key, value in choice.items():
                        choice_tuples.append((key, value))
                field_data['field_args']['choices'] = choice_tuples
            except KeyError:
                pass
            try:
                boolean_dicts = field_data['field_args'].pop('boolean_dicts')
                boolean_tuples = []
                for boolean in boolean_dicts:
                    for key, value in boolean.items():
                        boolean_tuples.append((key, value))
                field_data['field_args']['choices'] = boolean_tuples
                field_data['field_args']['coerce'] = lambda x: x == 'true'
            except KeyError:
                pass
            for key, value in field_data['field_args'].items():
                setattr(added_field, key, value)
            self.fields[field_data['field_name']] = added_field


class GeneratedFormPublic(GeneratedForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ''' Begin hidden field '''
        self.fields['email'] = EmailField(
            label='',
            required=False,
        )
        ''' End hidden field '''
        if hasattr(settings, 'RECAPTCHA_PUBLIC_KEY') and hasattr(settings, 'RECAPTCHA_PRIVATE_KEY'):
            try:
                from captcha.fields import (
                    ReCaptchaField,
                )
                self.fields['captcha'] = ReCaptchaField(
                    label='',
                    widget=get_recaptcha_widget(),
                )
            except ImportError:
                pass
